# Fremde Importe:

import re

from dataclasses import dataclass
from enum import Enum
from typing import Callable, Iterable, TypeVar


# Eigene Importe:

from . import color
from . import style

from .helpers import *


class Alignment(Enum):

    LEFT = 0
    RIGHT = 1
    PYSTD = 2


class InputStatus(Enum):

    COMMAND_ENTERED = 1
    ABORT_TRIGGERED = 2


@dataclass(kw_only=True)
class ModePreset:
    
    fg: color.Color | int | color.RGBColor | None = None
    bg: color.Color | int | color.RGBColor | None = None
    styles: Iterable[style.Style] | style.Style | None = None


@default_str_and_repr
class Command:
    
    def _init_(self, name: str, action: Callable[[list[str]], None]) -> None:
        self.name = name
        self.action = action


@dataclass(kw_only=True)
class InputMarkingTags:
    
    valid_input: str | None = None
    invalid_input: str | None = None
    command: str | None = None
    abort_string: str | None = None


class _BracesRepl:
    OPEN = '\x11'
    CLOSED = '\x12'


_T = TypeVar('_T')

_COLORS_FOR_KEYS = {
    'd': color.DEFAULT_COLOR,
    'n': color.BLACK,
    'r': color.RED,
    'g': color.GREEN,
    'y': color.YELLOW,
    'b': color.BLUE,
    'm': color.MAGENTA,
    'c': color.CYAN,
    'w': color.WHITE,
    'N': color.BBLACK,
    'R': color.BRED,
    'G': color.BGREEN,
    'Y': color.BYELLOW,
    'B': color.BBLUE,
    'M': color.BMAGENTA,
    'C': color.BCYAN,
    'W': color.BWHITE
}

_STYLES_FOR_KEYS = {
    'f': style.BOLD,
    'D': style.DIM,
    'i': style.ITALIC,
    'u': style.UNDERLINE,
    'U': style.DOUBLE_UNDERLINE,
    'o': style.BLINKING,
    'O': style.RAPID_BLINKING,
    'I': style.INVERSE,
    'x': style.HIDDEN,
    's': style.STRIKETHROUGH
}

_COLOR_KEYS = ''.join(_COLORS_FOR_KEYS)
_STYLE_KEYS = ''.join(_STYLES_FOR_KEYS)


# Öffentliche Funktionen:

def escseq(*, fg: color.Color | int | color.RGBColor | None = None, bg: color.Color | int | color.RGBColor | None = None,
           styles: Iterable[style.Style] | style.Style | None = None, mode_preset: ModePreset | None = None) -> str:
    codes = _code_list(styles=styles, bg=bg, fg=fg, mode_preset=mode_preset)
    return _escseq_for_codes(codes)


def highlight(value: object, *, fg: color.Color | int | color.RGBColor | None = None, bg: color.Color | int | color.RGBColor | None = None,
              styles: Iterable[style.Style] | style.Style | None = None, mode_preset: ModePreset | None = None) -> str:
    codes = _code_list(styles=styles, bg=bg, fg=fg, mode_preset=mode_preset)
    if len(codes) == 0:
        return str(value)
    escseq = _escseq_for_codes(codes)
    return f'{escseq}{value}{style.RESET_ALL.escseq}'


def hformat(format_str: str, tag_braces: str = '<>') -> str:
    if len(tag_braces) != 2:
        raise ValueError('tag_braces argument must be length 2')
    (otb, ctb) = tag_braces # open tag brace, closed tag brace
    format_str = format_str.replace(f'\\{otb}', _BracesRepl.OPEN)
    format_str = format_str.replace(f'\\{ctb}', _BracesRepl.CLOSED)
    tag_re = re.escape(otb) + r'.*?' + re.escape(ctb)
    tags = re.findall(tag_re, format_str)
    parts = re.split(tag_re, format_str)
    escseqs = [_tag_to_escseq(t) for t in tags]
    result = '%s%s%s' % (''.join(p + s for (p, s) in zip(parts, escseqs)), parts[-1], style.RESET_ALL.escseq)
    result = result.replace(_BracesRepl.OPEN, otb)
    result = result.replace(_BracesRepl.CLOSED, ctb)
    return result


def hprint(*values: object, fg: color.Color | int | color.RGBColor | None = None, bg: color.Color | int | color.RGBColor | None = None,
           styles: Iterable[style.Style] | style.Style | None = None, mode_preset: ModePreset | None = None, sep: str | None = ' ',
           end: str | None = '\n') -> None:
    print(escseq(fg=fg, bg=bg, styles=styles, mode_preset=mode_preset), end='')
    print(*values, sep=sep, end='')
    print(style.RESET_ALL.escseq, end=end)


def hprintf(format_str: str, tag_braces: str | None = '<>', end: str | None = '\n') -> str:
    print(hformat(format_str, tag_braces), end=end)


def hinput(prompt_format: str, input_tag: str, *, tag_braces: str = '<>', repeat_on_empty: bool = False, validate_func: Callable[[str], str | None] | None = None,
           commands: Iterable[Command] | Command | None = None, abort_strings: Iterable[str] | str | None = None,
           special_inputs_ignore_case: bool = False, input_marking_tags: InputMarkingTags | None = None) -> str | InputStatus:
    commands = _ensure_iterable(commands)
    abort_strings = _ensure_iterable(abort_strings)
    if '' in (cmd.name for cmd in commands) or '' in abort_strings:
        repeat_on_empty = False
    prompt = hformat(prompt_format, tag_braces) + _tag_to_escseq(input_tag)
    prompt_len = len(unhighlight(prompt))
    while True:
        inp = input(prompt)
        reset_all()
        if repeat_on_empty and len(inp) == 0:
            continue
        entered_command = _get_entered_command(inp, commands, special_inputs_ignore_case)
        if entered_command is not None:
            input_marking_tags and _mark_input(inp, prompt_len, input_marking_tags.command)
            args = inp.split()
            entered_command.action(args[1:])
            return InputStatus.COMMAND_ENTERED
        entered_abort_string = _get_entered_string(inp, abort_strings, special_inputs_ignore_case)
        if entered_abort_string is not None:
            input_marking_tags and _mark_input(inp, prompt_len, input_marking_tags.abort_string)
            return InputStatus.ABORT_TRIGGERED
        if validate_func:
            error_msg = validate_func(inp)
            if error_msg:
                input_marking_tags and _mark_input(inp, prompt_len, input_marking_tags.invalid_input)
                hprintf(error_msg)
                continue
        input_marking_tags and _mark_input(inp, prompt_len, input_marking_tags.valid_input)
        return inp


def line_breaks(n: int) -> None:
    for _ in range(n):
        print()


def reset_all() -> None:
    print(style.RESET_ALL.escseq, end='')


def unhighlight(s: str) -> str:
    return re.sub(r'\x1b\[.+?m', '', s)


def table_256_fg_colors(wide: bool = False, fillchar: str = ' ', bg_color: color.Color | int | color.RGBColor | None = None,
                        styles: Iterable[style.Style] | style.Style | None = None) -> str:
    return _table_256_colors(for_bg=False, wide=wide, fillchar=fillchar, other_color=bg_color, styles=styles)


def table_256_bg_colors(wide: bool = False, fg_color: color.Color | int | color.RGBColor | None = None,
                        styles: Iterable[style.Style] | style.Style | None = None) -> str:
    return _table_256_colors(for_bg=True, wide=wide, fillchar=' ', other_color=fg_color, styles=styles)


def move_cursor_up(lines: int) -> None:
    _move_cursor('A', lines)


def move_cursor_down(lines: int) -> None:
    _move_cursor('B', lines)


def move_cursor_right(places: int) -> None:
    _move_cursor('C', places)


def move_cursor_left(places: int) -> None:
    _move_cursor('D', places)


# Private Funktionen:

def _code_list(*, reset_all: bool = False, resets: list[style.Style] = [], styles: Iterable[style.Style] | style.Style | None = None,
                bg: color.Color | int | color.RGBColor | None = None, fg: color.Color | int | color.RGBColor | None = None, mode_preset: ModePreset | None = None) -> list[str]:
    (styles, bg, fg) = _apply_preset(styles, bg, fg, mode_preset)
    codes = [str(style.RESET_ALL.code)] if reset_all else []
    codes.extend([str(r.reset.code) for r in resets])
    codes.extend([str(s.code) for s in styles])
    if bg is not None:
        codes.append(_code_for_color(bg, for_bg=True))
    if fg is not None:
        codes.append(_code_for_color(fg, for_bg=False))
    return codes


def _apply_preset(styles: Iterable[style.Style] | style.Style | None, bg: color.Color | int | color.RGBColor | None, fg: color.Color | int | color.RGBColor | None,
        mode_preset: ModePreset | None) -> tuple[Iterable[style.Style], color.Color | int | color.RGBColor | None, color.Color | int | color.RGBColor | None]:
    styles = _ensure_iterable(styles)
    if mode_preset is None:
        return (styles, bg, fg)
    styles = styles if styles else mode_preset.styles
    bg = fg if bg is not None else mode_preset.bg
    fg = fg if fg is not None else mode_preset.fg
    return (styles, bg, fg)


def _code_for_color(clr: color.Color | int | color.RGBColor, *, for_bg: bool) -> str:
    if isinstance(clr, color.Color):
        return str((clr.bg if for_bg else clr.fg).code)
    if isinstance(clr, int):
        return f'{48 if for_bg else 38};5;{clr}'
    return clr.bg.code if for_bg else clr.fg.code


def _escseq_for_codes(codes: list[str]) -> str:
    if len(codes) == 0:
        return ''
    else:
        return '\x1b[%sm' % ';'.join(codes)


def _ensure_iterable(value: Iterable[_T] | _T | None) -> Iterable[_T]:
    if value is None:
        return ()
    if isinstance(value, Iterable):
        return value
    return (value,)


def _center_right_aligned(s: str, width: int, fillchar: str) -> str:
    spaces = (width - len(s))
    rspaces = spaces // 2
    lspaces = rspaces + (spaces % 2 != 0)
    return '%s%s%s' % (fillchar * lspaces, s, fillchar * rspaces)


def _tag_to_escseq(tag: str) -> str:
    content = tag[1:-1]
    reset_all = False
    resets: list[style.Style] = []
    styles: list[style.Style] = []
    bg_color: color.Color | None = None
    fg_color: color.Color | None = None
    bg_marker = False
    reset_marker = False
    for char in content:
        if char in ' ,;#!':
            _assert_bg_marker_not_set(bg_marker, char, tag)
            _assert_reset_marker_not_set(reset_marker, char, tag)
            if char == '#':
                bg_marker = True
            elif char == '!':
                reset_marker = True
        elif char == 'a':
            _assert_bg_marker_not_set(bg_marker, char, tag)
            if reset_marker:
                reset_all = True
                reset_marker = False
        elif char in _COLOR_KEYS:
            _assert_reset_marker_not_set(reset_marker, char, tag)
            clr = _COLORS_FOR_KEYS[char]
            if bg_marker:
                bg_color = clr
                bg_marker = False
            else:
                fg_color = clr
        elif char in _STYLE_KEYS:
            _assert_bg_marker_not_set(bg_marker, char, tag)
            stl = _STYLES_FOR_KEYS[char]
            if reset_marker:
                resets.append(stl)
                reset_marker = False
            else:
                styles.append(stl)
        else:
            raise _invalid_key_error(tag, char)
    return _escseq_for_tag(reset_all=reset_all, resets=resets, styles=styles, bg=bg_color, fg=fg_color)


def _assert_marker_not_set(marker: str, marker_set: bool, char: str, tag: str) -> None:
    if marker_set:
        raise _invalid_key_error(tag, f'{marker}{char}')
    

def _assert_bg_marker_not_set(marker_set: bool, char: str, tag: str) -> None:
    _assert_marker_not_set('#', marker_set, char, tag)
    

def _assert_reset_marker_not_set(marker_set: bool, char: str, tag: str) -> None:
    _assert_marker_not_set('!', marker_set, char, tag)


def _escseq_for_tag(reset_all: bool, resets: list[style.Style], styles: list[style.Style], bg: color.Color | None = None, fg: color.Color | None = None) -> str:
    codes = _code_list(reset_all=reset_all, resets=resets, styles=styles, bg=bg, fg=fg)
    return _escseq_for_codes(codes)


def _invalid_key_error(tag: str, invalid_key: str) -> ValueError:
    return ValueError(f'"{tag}": invalid format key: \'{invalid_key}\'')


def _move_cursor(direction: str, places: int) -> None:
    if places > 0:
        print(f'\x1b[{places}{direction}', end='')


# Private Hilfsfunktionen für die `fancyinput()`-Funktionen:

def _get_entered_string(inp: str, strings: Iterable[str], ignore_case: bool) -> str | None:
    inp = inp.split()[0]
    if ignore_case:
        inp = inp.casefold()
        strings = (v.casefold() for v in strings)
    return next((v for v in strings if v == inp), None)


def _get_entered_command(inp: str, commands: Iterable[Command], ignore_case: bool) -> Command | None:
    cmd_name = _get_entered_string(inp, (cmd.name for cmd in commands), ignore_case)
    return next((cmd for cmd in commands if cmd.name == cmd_name), None) if cmd_name else None


def _mark_input(inp: str, prompt_len: int, format_tag: str | None) -> None:
    if not format_tag:
        return
    move_cursor_up(1)
    move_cursor_right(prompt_len)
    hprintf(format_tag + inp)


# Private Hilfsfunktionen für die Funktionen `table_256_fg_colors()` und `table_256_bg_colors()`:

def _table_256_colors(for_bg: bool, wide: bool, fillchar: str, other_color: color.Color | int | color.RGBColor | None, styles: Iterable[style.Style] | style.Style | None) -> str:
    cw_216 = 5 if wide else 3 # cell width for 216 colors
    cw_top = cw_216 + 4 # cell width top (die ersten 16 Farben)
    cw_gs = cw_216 + 2 # well width grayscale

    # Alles soll zum Ergebnis-String hinzugefügt werden
    result = _table_256_colors_heading('Standard colors', cw_top * 8)
    result += ' ' * cw_top
    result += _table_256_colors_heading('High-intensity colors', cw_top * 8)
    result += '\n'

    # Farben von 0 bis 7 (8 Stk.) und von 8 bis 15 (8 Stk.)
    result += _table_256_colors_line(for_bg, 0, 8, cw_top, fillchar, other_color, styles)
    result += ' ' * cw_top
    result += _table_256_colors_line(for_bg, 8, 8, cw_top, fillchar, other_color, styles)
    result += '\n\n'
    
    # 216 Farben
    result += _table_256_colors_heading('216 colors', cw_216 * 36)
    result += '\n'
    for i in range(6):
        result += _table_256_colors_line(for_bg, 36 * i + 16, 36, cw_216, fillchar, other_color, styles)
        result += '\n'
    result += '\n'

    # Graustufen
    result += _table_256_colors_heading('Grayscale colors', cw_gs * 24)
    result += '\n'
    result += _table_256_colors_line(for_bg, 232, 24, cw_gs, fillchar, other_color, styles)

    # Rückgabe
    return result


def _table_256_colors_line(for_bg: bool, start_color_num: int, n_colors: int, cell_width: int, fillchar: str, other_color: color.Color | int | color.RGBColor | None,
                            styles: Iterable[style.Style] | style.Style | None) -> str:
    return ''.join([_table_256_colors_cell(for_bg, start_color_num + i, cell_width, fillchar, other_color, styles) for i in range(n_colors)])


def _table_256_colors_cell(for_bg: bool, color_num: int, width: int, fillchar: str, other_color: color.Color | int | color.RGBColor | None,
                            styles: Iterable[style.Style] | style.Style | None) -> str:
    (fg, bg) = (other_color, color_num) if for_bg else (color_num, other_color)
    return highlight(_center_right_aligned(str(color_num), width, fillchar), fg=fg, bg=bg, styles=styles)


def _table_256_colors_heading(heading: str, width: int) -> str:
    return highlight(_center_right_aligned(heading, width, ' '), fg=color.BWHITE, styles=[style.BOLD])